from typing import Optional, Union, Dict, Any
from ..utils._base_chain import Chain
from langchain_core.language_models.llms import LLM

from langchain_timbr.utils.timbr_utils import get_timbr_agent_options

from ..utils.general import parse_list, to_boolean, to_integer, validate_timbr_connection_params, sanitize_results
from ..utils.timbr_llm_utils import determine_concept
from ..llm_wrapper.llm_wrapper import LlmWrapper
from .. import config


class IdentifyTimbrConceptChain(Chain):
    """
    LangChain chain for identifying relevant concepts from user prompts using Timbr knowledge graphs.
    
    This chain analyzes natural language prompts to determine the most appropriate concept(s)
    within a Timbr ontology/knowledge graph that best matches the user's intent. It uses an LLM
    to process prompts and connects to Timbr via URL and token for concept identification.
    """

    _ontology: Optional[str] = None
    
    def __init__(
        self,
        llm: Optional[LLM] = None,
        url: Optional[str] = None,
        token: Optional[str] = None,
        ontology: Optional[str] = None,
        concepts_list: Optional[Union[list[str], str]] = None,
        views_list: Optional[Union[list[str], str]] = None,
        include_logic_concepts: Optional[bool] = False,
        include_tags: Optional[Union[list[str], str]] = None,
        should_validate: Optional[bool] = False,
        retries: Optional[int] = 3,
        note: Optional[str] = '',
        agent: Optional[str] = None,
        verify_ssl: Optional[bool] = True,
        is_jwt: Optional[bool] = False,
        jwt_tenant_id: Optional[str] = None,
        conn_params: Optional[dict] = None,
        debug: Optional[bool] = False,
        **kwargs,
    ):
        """
        :param llm: An LLM instance or a function that takes a prompt string and returns the LLM's response (optional, will use LlmWrapper with env variables if not provided)
        :param url: Timbr server url (optional, defaults to TIMBR_URL environment variable)
        :param token: Timbr password or token value (optional, defaults to TIMBR_TOKEN environment variable)
        :param ontology: The name of the ontology/knowledge graph (optional, defaults to ONTOLOGY/TIMBR_ONTOLOGY environment variable)
        :param concepts_list: Optional specific concept options to query
        :param views_list: Optional specific view options to query
        :param include_logic_concepts: Optional boolean to include logic concepts (concepts without unique properties which only inherits from an upper level concept with filter logic) in the query.
        :param include_tags: Optional specific concepts & properties tag options to use in the query (Disabled by default. Use '*' to enable all tags or a string represents a list of tags divided by commas (e.g. 'tag1,tag2')
        :param should_validate: Whether to validate the identified concept before returning it
        :param retries: Number of retry attempts if the identified concept is invalid
        :param note: Optional additional note to extend our llm prompt
        :param agent: Optional Timbr agent name for options setup.
        :param verify_ssl: Whether to verify SSL certificates (default is True).
        :param is_jwt: Whether to use JWT authentication (default is False).
        :param jwt_tenant_id: JWT tenant ID for multi-tenant environments (required when is_jwt=True).
        :param conn_params: Extra Timbr connection parameters sent with every request (e.g., 'x-api-impersonate-user').
        :param kwargs: Additional arguments to pass to the base
        
        ## Example
        ```
        # Using explicit parameters
        identify_timbr_concept_chain = IdentifyTimbrConceptChain(
            llm=<llm or timbr_llm_wrapper instance>,
            url=<url>,
            token=<token>,
            ontology=<ontology_name>,
            concepts_list=<concepts>,
            views_list=<views>,
            include_tags=<tags>,
            note=<note>,
        )

        # Using environment variables for timbr environment (TIMBR_URL, TIMBR_TOKEN, TIMBR_ONTOLOGY)
        identify_timbr_concept_chain = IdentifyTimbrConceptChain(
            llm=<llm or timbr_llm_wrapper instance>,
        )

        # Using environment variables for both timbr environment & llm (TIMBR_URL, TIMBR_TOKEN, TIMBR_ONTOLOGY, LLM_TYPE, LLM_API_KEY, etc.)
        identify_timbr_concept_chain = IdentifyTimbrConceptChain()

        return identify_timbr_concept_chain.invoke({ "prompt": question }).get("concept", None)
        ```
        """
        super().__init__(**kwargs)
        
        # Initialize LLM - use provided one or create with LlmWrapper from env variables
        if llm is not None:
            self._llm = llm
        else:
            try:
                self._llm = LlmWrapper()
            except Exception as e:
                raise ValueError(f"Failed to initialize LLM from environment variables. Either provide an llm parameter or ensure LLM_TYPE and LLM_API_KEY environment variables are set. Error: {e}")
        
        self._url = url if url is not None else config.url
        self._token = token if token is not None else config.token
        
        # Validate required parameters
        validate_timbr_connection_params(self._url, self._token)
        
        self._verify_ssl = to_boolean(verify_ssl)
        self._is_jwt = to_boolean(is_jwt)
        self._jwt_tenant_id = jwt_tenant_id
        self._debug = to_boolean(debug)
        self._conn_params = conn_params or {}

        self._agent = agent
        if self._agent:
            agent_options = get_timbr_agent_options(self._agent, conn_params=self._get_conn_params())

            self._ontology = agent_options.get("ontology") if "ontology" in agent_options else None
            self._concepts_list = parse_list(agent_options.get("concepts_list")) if "concepts_list" in agent_options else None
            self._views_list = parse_list(agent_options.get("views_list")) if "views_list" in agent_options else None
            self._include_logic_concepts = to_boolean(agent_options.get("include_logic_concepts")) if "include_logic_concepts" in agent_options else False
            self._include_tags = parse_list(agent_options.get("include_tags")) if "include_tags" in agent_options else None
            self._should_validate = to_boolean(agent_options.get("should_validate")) if "should_validate" in agent_options else False
            self._retries = to_integer(agent_options.get("retries") if "retries" in agent_options else retries)
            self._note = agent_options.get("note") if "note" in agent_options else ''
            if note:
                self._note = ((self._note + '\n') if self._note else '') + note
        else:
            self._ontology = ontology if ontology is not None else config.ontology
            self._concepts_list = parse_list(concepts_list)
            self._views_list = parse_list(views_list)
            self._include_logic_concepts = to_boolean(include_logic_concepts)
            self._include_tags = parse_list(include_tags)
            self._should_validate = to_boolean(should_validate)
            self._retries = to_integer(retries)
            self._note = note
        

    @property
    def usage_metadata_key(self) -> str:
        return "identify_concept_usage_metadata"


    @property
    def input_keys(self) -> list:
        return ["prompt"]


    @property
    def output_keys(self) -> list:
        base = [
            "ontology",
            "schema",
            "concept",
            "concept_metadata",
            "identify_concept_reason",
            self.usage_metadata_key]
        return list(dict.fromkeys(self.input_keys + base))


    def _get_conn_params(self) -> dict:
        return {
            "url": self._url,
            "token": self._token,
            "ontology": self._ontology if self._ontology is not None else config.ontology,
            "verify_ssl": self._verify_ssl,
            "is_jwt": self._is_jwt,
            "jwt_tenant_id": self._jwt_tenant_id,
            **self._conn_params,
        }


    def _call(self, inputs: Dict[str, Any], run_manager=None) -> Dict[str, str]:
        prompt = inputs["prompt"]
        res = determine_concept(
            question=prompt,
            llm=self._llm,
            conn_params=self._get_conn_params(),
            concepts_list=self._concepts_list,
            views_list=self._views_list,
            include_logic_concepts=self._include_logic_concepts,
            include_tags=self._include_tags,
            should_validate=self._should_validate,
            retries=self._retries,
            note=self._note,
            debug=self._debug,
        )

        usage_metadata = res.pop("usage_metadata", {})
        result = {
            **inputs,
            **res,
            self.usage_metadata_key: usage_metadata,
        }
        return sanitize_results(self.output_keys, result)



from typing import Any, Dict, List
from langchain_core.runnables import Runnable


class Chain(Runnable):
    """
    Compatibility base class that mimics the legacy langchain.chains.base.Chain
    interface (removed in langchain 1.x).

    Subclasses should implement:
      - ``_call(self, inputs, run_manager=None) -> dict``
      - ``input_keys`` property
      - ``output_keys`` property
    """

    def __init__(self, **kwargs):
        pass

    @property
    def input_keys(self) -> List[str]:
        return []

    @property
    def output_keys(self) -> List[str]:
        return []

    def _call(self, inputs: Dict[str, Any], run_manager=None) -> Dict[str, Any]:
        raise NotImplementedError

    def invoke(self, input: Dict[str, Any], config=None, **kwargs) -> Dict[str, Any]:
        return self._call(input)

    async def ainvoke(self, input: Dict[str, Any], config=None, **kwargs) -> Dict[str, Any]:
        return self._call(input)



from typing import Optional, Any, Literal
from typing_extensions import TypedDict
from langchain_core.language_models.llms import LLM
from langgraph.graph import StateGraph, END

from .utils.general import to_boolean, to_integer
from .llm_wrapper.llm_wrapper import LlmWrapper
from .utils.timbr_utils import get_ontologies, get_concepts
from .langchain import IdentifyTimbrConceptChain, GenerateTimbrSqlChain, ValidateTimbrSqlChain, ExecuteTimbrQueryChain, create_timbr_sql_agent
from .langgraph import GenerateTimbrSqlNode, ValidateSemanticSqlNode, ExecuteSemanticQueryNode, GenerateResponseNode


from . import config

class TimbrLanggraphState(TypedDict):
    prompt: str
    sql: str
    concept: str
    rows: list
    response: str
    error: str
    is_sql_valid: bool
    usage_metadata: dict[str, Any]


class TimbrLlmConnector:
    def __init__(
        self,
        llm: LLM,
        url: Optional[str] = config.url,
        token: Optional[str] = config.token,
        ontology: Optional[str] = config.ontology,
        max_limit: Optional[int] = config.llm_default_limit,
        verify_ssl: Optional[bool] = True,
        is_jwt: Optional[bool] = False,
        jwt_tenant_id: Optional[str] = None,
        conn_params: Optional[dict] = None,
    ):
        """
        :param url: Timbr server url
        :param token: Timbr password or token value
        :param ontology: The name of the ontology/knowledge graph
        :param llm: An LLM instance or a function that takes a prompt string and returns the LLM’s response
        :param max_limit: Maximum number of rows to return
        :param verify_ssl: Whether to verify SSL certificates (default is True).
        :param is_jwt: Whether to use JWT authentication (default is False).
        :param jwt_tenant_id: Tenant ID for JWT authentication (if applicable).
        :param conn_params: Extra Timbr connection parameters sent with every request (e.g., 'x-api-impersonate-user').

        ## Example
        ```
        timbr_llm_wrapper = LlmWrapper(
            llm_type=LlmTypes.OpenAI,
            model="gpt-4o"
            api_key=<openai_api_key>
        )

        llm_connector = TimbrLlmConnector(
            url=<url>,
            token=<token>,
            llm=timbr_llm_wrapper,
        )

        # Show ontology list at timbr instance from url connection
        ontologies = llm_connector.get_ontologies()

        # Find which concept & schema will be queried by the user input
        determine_concept_res = llm_connector.determine_concept(llm_input)
        query_concept, query_schema = determine_concept_res.get("concept"), determine_concept_res.get("schema")

        # Generate timbr SQL query from user input
        sql_query = llm_connector.generate_sql(llm_input).get("sql")

        # Run timbr SQL query
        results = llm_connector.run_timbr_query(sql_query).get("rows", [])
        
        # Parse & Run LLM question
        results = llm_connector.run_llm_query(llm_input).get("rows", [])
        ```
        """
        self.url = url
        self.token = token
        self.ontology = ontology
        self.max_limit = to_integer(max_limit)
        self.verify_ssl = to_boolean(verify_ssl)
        self.is_jwt = to_boolean(is_jwt)
        self.jwt_tenant_id = jwt_tenant_id
        self.conn_params = conn_params or {}
        
        if llm is not None:
            self._llm = llm
        elif config.llm_type is not None and config.llm_api_key is not None:
            llm_params = {}
            if config.llm_temperature is not None:
                llm_params["temperature"] = config.llm_temperature

            self._llm = LlmWrapper(
                llm_type=config.llm_type,
                api_key=config.llm_api_key,
                model=config.llm_model,
                **llm_params,
            )


    # TODO: Make this function a decorator and use in on relevant methods
    # def _is_ontology_set(self):
    #     return self.ontology != 'system_db'
    

    def _get_conn_params(self):
        return {
            "url": self.url,
            "token": self.token,
            "ontology": self.ontology,
            "verify_ssl": self.verify_ssl,
            "is_jwt": self.is_jwt,
            "jwt_tenant_id": self.jwt_tenant_id,
            "additional_headers": {"results-limit": str(self.max_limit)},
            **self.conn_params,
        }


    def get_ontologies(self) -> list[str]:
        return get_ontologies(conn_params=self._get_conn_params())


    def get_concepts(self) -> dict:
        """
        Get the list of concepts from the Timbr server.
        """
        return get_concepts(
            conn_params=self._get_conn_params(),
            concepts_list="*",
        )
    

    def get_views(self) -> dict:
        """
        Get the list of views from the Timbr server.
        """
        return get_concepts(
            conn_params=self._get_conn_params(),
            views_list="*",
        )


    def set_ontology(self, ontology: str):
        self.ontology = ontology


    def determine_concept(
        self,
        question: str,
        concepts_list: Optional[list] = None,
        views_list: Optional[list] = None,
        include_logic_concepts: Optional[bool] = False,
        include_tags: Optional[str] = None,
        should_validate: Optional[bool] = False,
        retries: Optional[int] = 3,
        note: Optional[str] = '',
        **chain_kwargs: Any,
    ) -> dict[str, Any]:
        determine_concept_chain = IdentifyTimbrConceptChain(
            **self._get_conn_params(),
            llm=self._llm,
            concepts_list=concepts_list,
            views_list=views_list,
            include_logic_concepts=include_logic_concepts,
            include_tags=include_tags,
            should_validate=should_validate,
            retries=retries,
            note=note,
            **chain_kwargs,
        )

        return determine_concept_chain.invoke({ "prompt": question })


    def generate_sql(
        self,
        question: str,
        concept_name: Optional[str] = None,
        schema: Optional[str] = None,
        concepts_list: Optional[list] = None,
        views_list: Optional[list] = None,
        include_logic_concepts: Optional[bool] = False,
        include_tags: Optional[str] = None,
        should_validate_sql: Optional[bool] = config.should_validate_sql,
        retries: Optional[int] = 3,
        note: Optional[str] = '',
        **chain_kwargs: Any,
    ) -> dict[str, Any]:
        generate_timbr_llm_chain = GenerateTimbrSqlChain(
            llm=self._llm,
            **self._get_conn_params(),
            schema=schema,
            concept=concept_name,
            concepts_list=concepts_list,
            views_list=views_list,
            include_logic_concepts=include_logic_concepts,
            include_tags=include_tags,
            should_validate_sql=should_validate_sql,
            retries=retries,
            max_limit=self.max_limit,
            note=note,
            **chain_kwargs,
        )

        return generate_timbr_llm_chain.invoke({ "prompt": question })


    def validate_sql(
        self,
        question: str,
        sql_query: str,
        retries: Optional[int] = 3,
        concepts_list: Optional[list] = None,
        views_list: Optional[list] = None,
        include_logic_concepts: Optional[bool] = False,
        include_tags: Optional[str] = None,
        note: Optional[str] = '',
        **chain_kwargs: Any,
    ) -> dict[str, Any]:
        validate_timbr_sql_chain = ValidateTimbrSqlChain(
            llm=self._llm,
            **self._get_conn_params(),
            retries=retries,
            concepts_list=concepts_list,
            views_list=views_list,
            include_logic_concepts=include_logic_concepts,
            include_tags=include_tags,
            max_limit=self.max_limit,
            note=note,
            **chain_kwargs,
        )
        return validate_timbr_sql_chain.invoke({ "sql": sql_query, "prompt": question })


    def run_timbr_query(
        self,
        sql_query: str,
        concepts_list: Optional[list] = None,
        views_list: Optional[list] = None,
        include_logic_concepts: Optional[bool] = False,
        include_tags: Optional[str] = None,
        should_validate_sql: Optional[bool] = config.should_validate_sql,
        retries: Optional[int] = 3,
        note: Optional[str] = '',
        **chain_kwargs: Any,
    ) -> dict[str, Any]:
        execute_timbr_query_chain = ExecuteTimbrQueryChain(
            llm=self._llm,
            **self._get_conn_params(),
            concepts_list=concepts_list,
            views_list=views_list,
            include_logic_concepts=include_logic_concepts,
            include_tags=include_tags,
            should_validate_sql=should_validate_sql,
            retries=retries,
            max_limit=self.max_limit,
            note=note,
            **chain_kwargs,
        )

        return execute_timbr_query_chain.invoke({ "sql": sql_query })


    def run_llm_query(
        self,
        question: str,
        concepts_list: Optional[list] = None,
        views_list: Optional[list] = None,
        include_logic_concepts: Optional[bool] = False,
        include_tags: Optional[str] = None,
        should_validate_sql: Optional[bool] = config.should_validate_sql,
        retries: Optional[int] = 3,
        note: Optional[str] = '',
        **agent_kwargs: Any,
    ) -> dict[str, Any]:
        agent = create_timbr_sql_agent(
            llm=self._llm,
            conn_params=self._get_conn_params(),
            concept=None,
            concepts_list=concepts_list,
            views_list=views_list,
            include_logic_concepts=include_logic_concepts,
            include_tags=include_tags,
            should_validate_sql=should_validate_sql,
            retries=retries,
            max_limit=self.max_limit,
            note=note,
            **agent_kwargs,
        )

        return agent.invoke(question)


    def run_llm_query_graph(
        self,
        question: str,
        concepts_list: Optional[list] = None,
        views_list: Optional[list] = None,
        include_logic_concepts: Optional[bool] = False,
        include_tags: Optional[str] = None,
        should_validate_sql: Optional[bool] = config.should_validate_sql,
        retries: Optional[int] = 3,
        note: Optional[str] = '',
        **nodes_kwargs: Any,
    ) -> dict[str, Any]:
        generate_sql_node = GenerateTimbrSqlNode(
            llm=self._llm,
            **self._get_conn_params(),
            concepts_list=concepts_list,
            views_list=views_list,
            include_logic_concepts=include_logic_concepts,
            include_tags=include_tags,
            max_limit=self.max_limit,
            note=note,
            **nodes_kwargs,
        )
        validate_sql_node = ValidateSemanticSqlNode(
            llm=self._llm,
            **self._get_conn_params(),
            retries=retries,
            concepts_list=concepts_list,
            views_list=views_list,
            include_logic_concepts=include_logic_concepts,
            include_tags=include_tags,
            max_limit=self.max_limit,
            note=note,
            **nodes_kwargs,
        )
        execute_sql_node = ExecuteSemanticQueryNode(
            llm=self._llm,
            **self._get_conn_params(),
            concepts_list=concepts_list,
            views_list=views_list,
            include_logic_concepts=include_logic_concepts,
            include_tags=include_tags,
            should_validate_sql=should_validate_sql,
            retries=retries,
            max_limit=self.max_limit,
            note=note,
            **nodes_kwargs,
        )
        generate_response_node = GenerateResponseNode()

        graph_builder = StateGraph(TimbrLanggraphState)

        graph_builder.add_node("generate_sql", generate_sql_node)
        graph_builder.add_node("validate_sql", validate_sql_node)
        graph_builder.add_node("execute_sql", execute_sql_node)
        graph_builder.add_node("generate_response", generate_response_node)

        graph_builder.add_edge("generate_sql", "validate_sql")
        
        def route_validation(state: dict) -> Literal["execute_sql", "end"]:
            # If validation is successful, proceed to execute the query.
            # Otherwise, stop the flow.
            if state.get("is_sql_valid"):
                return "execute_sql"
            else:
                return "end"
            
        graph_builder.add_conditional_edges(
            "validate_sql", 
            route_validation,
            {
                "execute_sql": "execute_sql",
                "end": END
            }
        )
        
        graph_builder.add_edge("execute_sql", "generate_response")
        graph_builder.set_entry_point("generate_sql")

        compiled_graph = graph_builder.compile()
        
        initial_state = {
            "prompt": question,
            "sql": "",
            "concept": "",
            "rows": [],
            "response": "",
            "error": "",
            "is_sql_valid": False,
            "usage_metadata": {}
        }
        
        result = compiled_graph.invoke(initial_state)
        return result



import pytest

from langchain_timbr import (
    IdentifyTimbrConceptChain,
    GenerateTimbrSqlChain,
    ValidateTimbrSqlChain,
    ExecuteTimbrQueryChain,
    GenerateAnswerChain,
    generate_key,
    decrypt_prompt,
)


class TestIdentifyTimbrConceptChain:
    """Test suite for IdentifyTimbrConceptChain functionality."""
    
    def test_identify_concept_chain(self, llm, config):
        """Test basic concept identification functionality."""
        chain = IdentifyTimbrConceptChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            verify_ssl=config["verify_ssl"],
        )
        result = chain.invoke({ "prompt": config["test_prompt"] })
        print("IdentifyTimbrConceptChain result:", result)
        assert "concept" in result, "Chain should return a 'concept'"
        assert result["concept"], "Returned concept should not be empty"
        assert "identify_concept_reason" in result, "Chain should return a 'identify_concept_reason'"
        assert result["identify_concept_reason"], "Returned identify_concept_reason should not be empty"
        assert chain.usage_metadata_key in result, "Chain should return 'usage_metadata'"
        assert len(result[chain.usage_metadata_key]) == 1 and 'determine_concept' in result[chain.usage_metadata_key], "Usage metadata should contain only 'determine_concept'"

    def test_identify_concept_chain_none_lists(self, llm, config):
        """Test concept identification with empty concept/view lists raises exception."""
        chain = IdentifyTimbrConceptChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            concepts_list='none',
            views_list='null',
            verify_ssl=config["verify_ssl"],
        )
        with pytest.raises(Exception, match="No relevant concepts found for the query"):
            result = chain.invoke({ "prompt": config["test_prompt"] })


class TestGenerateTimbrSqlChain:
    """Test suite for GenerateTimbrSqlChain functionality."""
    
    def test_generate_timbr_sql_chain(self, llm, config):
        """Test basic SQL generation functionality."""
        chain = GenerateTimbrSqlChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            verify_ssl=config["verify_ssl"],
        )
        result = chain.invoke({ "prompt": config["test_prompt"] })
        print("GenerateTimbrSqlChain result:", result)
        assert "sql" in result and result["sql"], "SQL should be generated"
        assert "concept" in result and result["concept"], "Concept name should be returned"
        assert chain.usage_metadata_key in result, "Chain should return 'usage_metadata'"
        assert len(result[chain.usage_metadata_key]) == 2 and 'determine_concept' in result[chain.usage_metadata_key] and 'generate_sql' in result[chain.usage_metadata_key], "Usage metadata should contain both 'determine_concept' and 'generate_sql'"
        assert "identify_concept_reason" in result, "Chain should return a 'identify_concept_reason'"
        assert result["identify_concept_reason"], "Returned identify_concept_reason should not be empty"
        assert "generate_sql_reason" in result, "Chain should return a 'generate_sql_reason'"
        assert result["generate_sql_reason"], "Returned generate_sql_reason should not be empty"


    def test_generate_timbr_sql_with_limit_chain(self, llm, config):
        """Test SQL generation with row limit."""
        chain = GenerateTimbrSqlChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            max_limit=2,
            verify_ssl=config["verify_ssl"],
        )
        result = chain.invoke({ "prompt": config["test_prompt"] })
        print("GenerateTimbrSqlChain with max_limit=2 result:", result)
        assert "sql" in result and result["sql"], "SQL should be generated"
        assert "LIMIT 2" in result["sql"].upper(), "SQL should contain 'LIMIT 2'"

    def test_generate_timbr_sql_with_concepts_list_chain(self, llm, config):
        """Test SQL generation with specific concepts list."""
        chain = GenerateTimbrSqlChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            concepts_list=['plant'],
            verify_ssl=config["verify_ssl"],
        )
        result = chain.invoke({ "prompt": config["test_prompt"] })
        print("GenerateTimbrSqlChain with concepts_list=['plant'] result:", result)
        assert "sql" in result and result["sql"], "SQL should be generated"
        assert "`dtimbr`.`plant`" in result["sql"], "SQL should reference '`dtimbr`.`plant`'"
        assert "concept" in result and result["concept"], "Concept name should be returned"
        assert result["concept"] == "plant", "Concept should equal 'plant'"

    def test_generate_timbr_sql_excluding_builtin_columns(self, llm, config):
        """Test SQL generation excluding built-in columns."""
        chain = GenerateTimbrSqlChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            max_limit=1,
            verify_ssl=config["verify_ssl"],
        )
        result = chain.invoke({ "prompt": config["test_prompt_2"] })
        print("GenerateTimbrSqlChain excluding built-in columns result:", result)
        assert "sql" in result and result["sql"], "SQL should be generated"
        builtin_columns = ["entity_type", "entity_id", "entity_label"]
        for col in builtin_columns:
            assert col not in result["sql"].lower(), f"SQL should not contain '{col}'"
        assert "concept" in result and result["concept"], "Concept name should be returned"

    def test_generate_timbr_sql_excluding_selected_columns(self, llm, config):
        """Test SQL generation excluding custom selected columns."""
        exclude_properties = ['customer_name', 'customer_segment']
        chain = GenerateTimbrSqlChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            exclude_properties=exclude_properties,
            max_limit=1,
            verify_ssl=config["verify_ssl"],
        )
        result = chain.invoke({ "prompt": config["test_prompt_2"] })
        print("GenerateTimbrSqlChain excluding selected columns result:", result)
        assert "sql" in result and result["sql"], "SQL should be generated"
            
        for col in exclude_properties:
            assert col not in result["sql"].lower(), f"SQL should not contain '{col}'"
        assert "concept" in result and result["concept"], "Concept name should be returned"

    def test_generate_timbr_sql_chain_descriptions(self, llm, config):
        """Test SQL generation with concept descriptions in debug mode."""
        chain = GenerateTimbrSqlChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            verify_ssl=config["verify_ssl"],
            debug=True
        )
        result = chain.invoke({ "prompt": config["test_prompt_2"] })
        print("GenerateTimbrSqlChain result:", result)
        assert "sql" in result and result["sql"], "SQL should be generated"
        assert "concept" in result and result["concept"] == "customer", "Concept customer should be returned"
        prompt = decrypt_prompt(result["generate_sql_usage_metadata"]["generate_sql"]["p_hash"], generate_key())
        assert "customer related info" in prompt, "Customer description should be in prompt"
        assert "concat of first and last name" in prompt, "Customer name description should be in prompt"
        assert "continent name" in prompt, "Order market description should be in prompt"
        assert "shipment of customer" in prompt, "Shipment relationship description should be in prompt"
        assert "count first names of customers" in prompt, "Customer first name count measure description should be in prompt"
        assert "calculation of revenue based on sales" in prompt, "Revenue measure description should be in prompt"

    def test_generate_timbr_sql_chain_descriptions_specific_concept(self, llm, config):
        """Test SQL generation with specific concept and descriptions."""
        chain = GenerateTimbrSqlChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            verify_ssl=config["verify_ssl"],
            concept="customer",
            debug=True
        )
        result = chain.invoke({ "prompt": config["test_prompt"] })
        print("GenerateTimbrSqlChain result:", result)
        assert "sql" in result and result["sql"], "SQL should be generated"
        assert "concept" in result and result["concept"] == "customer", "Concept customer should be returned"
        prompt = decrypt_prompt(result["generate_sql_usage_metadata"]["generate_sql"]["p_hash"], generate_key())
        assert "customer related info" in prompt, "Customer description should be in prompt"
        assert "concat of first and last name" in prompt, "Customer name description should be in prompt"
        assert "continent name" in prompt, "Order market description should be in prompt"
        assert "shipment of customer" in prompt, "Shipment relationship description should be in prompt"
        assert "count first names of customers" in prompt, "Customer first name count measure description should be in prompt"
        assert "calculation of revenue based on sales" in prompt, "Revenue measure description should be in prompt"

    def test_generate_timbr_sql_chain_descriptions_specific_cube(self, llm, config):
        """Test SQL generation with specific cube/view and descriptions."""
        chain = GenerateTimbrSqlChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            verify_ssl=config["verify_ssl"],
            views_list=["customer_cube"],
            debug=True
        )
        result = chain.invoke({ "prompt": config["test_prompt"] })
        print("GenerateTimbrSqlChain result:", result)
        assert "sql" in result and result["sql"], "SQL should be generated"
        assert "concept" in result and result["concept"] == "customer_cube", "Concept customer_cube should be returned"
        prompt = decrypt_prompt(result["generate_sql_usage_metadata"]["generate_sql"]["p_hash"], generate_key())
        assert "customer cube related info" in prompt, "Customer description should be in prompt"
        assert "concat of first and last name" in prompt, "Customer name description should be in prompt"
        assert "continent name" in prompt, "Order market description should be in prompt"
        assert "count first names of customers" in prompt, "Customer first name count measure description should be in prompt"
        assert "calculation of revenue based on sales" in prompt, "Revenue measure description should be in prompt"

    def test_generate_timbr_sql_chain_tags(self, llm, config):
        """Test SQL generation with tags filtering."""
        chain = GenerateTimbrSqlChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            verify_ssl=config["verify_ssl"],
            include_tags=["length", "synonym"],
            debug=True
        )
        result = chain.invoke({ "prompt": config["test_prompt_3"] })
        print("GenerateTimbrSqlChain result:", result)
        assert "sql" in result and result["sql"], "SQL should be generated"
        assert "concept" in result and result["concept"] == "product", "Concept product should be returned"
        assert chain.usage_metadata_key in result, "Chain should return 'usage_metadata'"
        prompt = decrypt_prompt(result["generate_sql_usage_metadata"]["generate_sql"]["p_hash"], generate_key())
        assert "commodity" in prompt, "Product tag value of synonym should be in prompt"
        assert "synonym" in prompt, "Product tag synonym should be in prompt"
        assert "length" in prompt, "Material tag should be in prompt"
        assert "max length 10 characters" in prompt, "Material tag value should be in prompt"
        assert "count of commodities" in prompt, "Product count measure tag value should be in prompt"

    def test_generate_timbr_sql_chain_tags_specific_concept(self, llm, config):
        """Test SQL generation with tags filtering and specific concept."""
        chain = GenerateTimbrSqlChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            verify_ssl=config["verify_ssl"],
            concept="product",
            include_tags=["length", "synonym"],
            debug=True
        )
        result = chain.invoke({ "prompt": config["test_prompt"] })
        print("GenerateTimbrSqlChain result:", result)
        assert "sql" in result and result["sql"], "SQL should be generated"
        assert "concept" in result and result["concept"] == "product", "Concept product should be returned"
        assert chain.usage_metadata_key in result, "Chain should return 'usage_metadata'"
        prompt = decrypt_prompt(result["generate_sql_usage_metadata"]["generate_sql"]["p_hash"], generate_key())
        assert "commodity" in prompt, "Product tag value of synonym should be in prompt"
        assert "synonym" in prompt, "Product tag synonym should be in prompt"
        assert "length" in prompt, "Material tag should be in prompt"
        assert "max length 10 characters" in prompt, "Material tag value should be in prompt"
        assert "count of commodities" in prompt, "Product count measure tag value should be in prompt"
        assert "total product price" in prompt, "Product total price measure tag value should be in prompt"

    def test_generate_timbr_sql_chain_tags_specific_cube(self, llm, config):
        """Test SQL generation with tags filtering and specific cube/view."""
        chain = GenerateTimbrSqlChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            verify_ssl=config["verify_ssl"],
            views_list=["product_cube"],
            include_tags=["length", "synonym"],
            debug=True
        )
        result = chain.invoke({ "prompt": config["test_prompt"] })
        print("GenerateTimbrSqlChain result:", result)
        assert "sql" in result and result["sql"], "SQL should be generated"
        assert "concept" in result and result["concept"] == "product_cube", "Concept product_cube should be returned"
        assert chain.usage_metadata_key in result, "Chain should return 'usage_metadata'"
        prompt = decrypt_prompt(result["generate_sql_usage_metadata"]["generate_sql"]["p_hash"], generate_key())
        assert "commodity cube" in prompt, "Product tag value of synonym should be in prompt"
        assert "synonym" in prompt, "Product tag synonym should be in prompt"
        assert "length" in prompt, "Material tag should be in prompt"
        assert "max length 10 characters" in prompt, "Material tag value should be in prompt"
        assert "count of commodities" in prompt, "Product count measure tag value should be in prompt"
        assert "total product price" in prompt, "Product total price measure tag value should be in prompt"


class TestValidateTimbrSqlChain:
    """Test suite for ValidateTimbrSqlChain functionality."""
    
    def test_validate_timbr_sql_chain(self, llm, config):
        """Test SQL validation functionality."""
        chain = ValidateTimbrSqlChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            retries=1,  # Use a single retry for test speed
            verify_ssl=config["verify_ssl"],
        )
        inputs = {
            "prompt": config["test_prompt"],
            "sql": "SELECT * FROM invalid_table",  # Intentionally invalid (or test with a valid one if available)
        }
        result = chain.invoke(inputs)
        print("ValidateTimbrSqlChain result:", result)
        # Check that we have a boolean flag and an error message (or None)
        assert "is_sql_valid" in result
        assert isinstance(result["is_sql_valid"], bool)
        # Optionally, assert that for a valid SQL the flag is True
        if result["sql"] and "invalid_table" not in result["sql"]:
            assert result["is_sql_valid"] is True


class TestExecuteTimbrQueryChain:
    """Test suite for ExecuteTimbrQueryChain functionality."""
    
    def test_execute_timbr_query_chain(self, llm, config):
        """Test basic query execution functionality."""
        chain = ExecuteTimbrQueryChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            verify_ssl=config["verify_ssl"],
        )
        # For testing, use a prompt and SQL that are expected to produce results.
        inputs = {
            "prompt": config["test_prompt"],
        }
        result = chain.invoke(inputs)
        print("ExecuteTimbrQueryChain result:", result)
        assert "rows" in result, "Result should contain 'rows'"
        assert isinstance(result["rows"], list), "'rows' should be a list"
        assert result["sql"], "SQL should be present in the result"

    def test_execute_timbr_query_chain_with_limit(self, llm, config):
        """Test query execution with row limit and tags."""
        chain = ExecuteTimbrQueryChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            max_limit=3,
            include_tags="*",
            verify_ssl=config["verify_ssl"],
        )
        inputs = {
            "prompt": config["test_prompt"],
        }
        result = chain.invoke(inputs)
        print("ExecuteTimbrQueryChain with max_limit=3 result:", result)
        assert "rows" in result, "Result should contain 'rows'"
        assert isinstance(result["rows"], list), "'rows' should be a list"
        assert result["sql"], "SQL should be present in the result"
        assert "LIMIT 3" in result["sql"].upper(), "SQL should contain 'LIMIT 3'"
        assert len(result["rows"]) <= 3, "Number of rows should not exceed max_limit"

    def test_execute_timbr_query_chain_with_result_inference(self, llm, config):
        """Test query execution with result inference and retry logic."""
        chain = ExecuteTimbrQueryChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"],
            retry_if_no_results=True,
            verify_ssl=config["verify_ssl"],
        )
        inputs = {
            "prompt": config["test_prompt"],
            "concept": "order",
            "schema": "dtimbr",
            "sql": "SELECT COUNT(1) AS `total_sales` FROM `dtimbr`.`order` WHERE order_city = 'Rishon LeZion'",
        }
        result = chain.invoke(inputs)
        print("ExecuteTimbrQueryChain with result inference result:", result)
        assert "rows" in result, "Result should contain 'rows'"
        assert isinstance(result["rows"], list), "'rows' should be a list"
        assert result["sql"], "SQL should be present in the result"
        assert "total_sales" in result["rows"][0], "Result rows should contain 'total_sales'"


class TestGenerateAnswerChain:
    """Test suite for GenerateAnswerChain functionality."""
    
    def test_generate_answer_chain(self, llm, config):
        """Test basic answer generation functionality."""
        chain = GenerateAnswerChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
        )
        result = chain.invoke({ "prompt": config["test_prompt"], "rows": [{ "total_sales": 1000 }] })
        print("GenerateAnswerChain result:", result)
        assert "answer" in result, "Chain should return an 'answer'"
        assert result["answer"], "Answer should not be empty"

    def test_generate_answer_chain_with_sql(self, llm, config):
        """Test answer generation with SQL context."""
        chain = GenerateAnswerChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
        )
        inputs = {
            "prompt": config["test_prompt"],
            "rows": [{ "counter": 1000 }],
            "sql": "SELECT COUNT(1) AS `counter` FROM `dtimbr`.`order` WHERE order_city = 'Rishon LeZion'",
        }
        result = chain.invoke(inputs)
        print("GenerateAnswerChain with SQL result:", result)
        assert "answer" in result, "Chain should return an 'answer'"
        assert result["answer"], "Answer should not be empty"
        assert chain.usage_metadata_key in result, "Chain should return 'generate_answer_usage_metadata'"
        assert len(result[chain.usage_metadata_key]) == 1 and 'answer_question' in result[chain.usage_metadata_key], "Generate answer chain usage metadata should contain only 'answer_question'"

    def test_generate_timbr_sql_with_no_dtimbr_perms(self, llm, config):
        """Test SQL generation with a user that lacks dtimbr schema permissions."""
        chain = GenerateTimbrSqlChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token_no_dtimbr_perms"],
            ontology=config["timbr_ontology_no_dtimbr_perms"],
            verify_ssl=config["verify_ssl"],
        )
        result = chain.invoke({ "prompt": "all calls" })
        print("GenerateTimbrSqlChain with concepts_list=['plant'] result:", result)
        assert "error" in result and "User doesn't have access to query Knowledge Graph schema: dtimbr" in result["error"], "Should return permission error message"


    def test_execute_with_multiple_ontologies(self, llm, config):
        """Test query execution with multiple ontologies"""
        chain = ExecuteTimbrQueryChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"] + ",timbr_patient_journey",
            max_limit=1,
            include_tags="*",
            verify_ssl=config["verify_ssl"],
        )
        inputs = {
            "prompt": "count number of customers",
        }
        result = chain.invoke(inputs)
        print("ExecuteTimbrQueryChain with count customers result:", result)
        assert "rows" in result, "Result should contain 'rows'"
        assert isinstance(result["rows"], list), "'rows' should be a list"
        assert result["sql"], "SQL should be present in the result"
        assert "CUSTOMER" in result["sql"].upper(), "SQL should contain 'CUSTOMER'"
        assert list(result["rows"][0].values())[0] == 781 or list(result["rows"][0].values())[0] == 20652, "Number of rows should not exceed max_limit"
        inputs = {
            "prompt": "count number of patients",
        }
        result = chain.invoke(inputs)
        print("ExecuteTimbrQueryChain with count patients result:", result)
        assert "rows" in result, "Result should contain 'rows'"
        assert isinstance(result["rows"], list), "'rows' should be a list"
        assert result["sql"], "SQL should be present in the result"
        assert "PATIENT" in result["sql"].upper(), "SQL should contain 'PATIENT'"
        assert list(result["rows"][0].values())[0] == 1176837, "Number of rows should not exceed max_limit"

    def test_execute_with_multiple_ontologies2(self, llm, config):
        """Test query execution with multiple ontologies."""
        chain = ExecuteTimbrQueryChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=config["timbr_ontology"] + ",timbr_patient_journey",
            concepts_list=["customer", "patient"],
            include_tags="*",
            verify_ssl=config["verify_ssl"],
        )
        inputs = {
            "prompt": "count number of customers",
        }
        result = chain.invoke(inputs)
        print("ExecuteTimbrQueryChain with count customers result:", result)
        assert "rows" in result, "Result should contain 'rows'"
        assert isinstance(result["rows"], list), "'rows' should be a list"
        assert result["sql"], "SQL should be present in the result"
        assert "CUSTOMER" in result["sql"].upper(), "SQL should contain 'CUSTOMER'"
        assert list(result["rows"][0].values())[0] == 781 or list(result["rows"][0].values())[0] == 20652, "Number of rows should not exceed max_limit"
        inputs = {
            "prompt": "count number of patients",
        }
        result = chain.invoke(inputs)
        print("ExecuteTimbrQueryChain with count patients result:", result)
        assert "rows" in result, "Result should contain 'rows'"
        assert isinstance(result["rows"], list), "'rows' should be a list"
        assert result["sql"], "SQL should be present in the result"
        assert "PATIENT" in result["sql"].upper(), "SQL should contain 'PATIENT'"
        assert list(result["rows"][0].values())[0] == 1176837, "Number of rows should not exceed max_limit"

    def test_execute_with_multiple_ontologies3(self, llm, config):
        """Test query execution with multiple ontologies."""
        chain = ExecuteTimbrQueryChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=[config["timbr_ontology"] , "timbr_patient_journey"],
            concepts_list=[config["timbr_ontology"] + "." + "customer", "timbr_patient_journey.patient"],
            include_tags="*",
            verify_ssl=config["verify_ssl"],
        )
        inputs = {
            "prompt": "count number of customers",
        }
        result = chain.invoke(inputs)
        print("ExecuteTimbrQueryChain with count customers result:", result)
        assert "rows" in result, "Result should contain 'rows'"
        assert isinstance(result["rows"], list), "'rows' should be a list"
        assert result["sql"], "SQL should be present in the result"
        assert "CUSTOMER" in result["sql"].upper(), "SQL should contain 'CUSTOMER'"
        assert list(result["rows"][0].values())[0] == 781 or list(result["rows"][0].values())[0] == 20652, "Number of rows should not exceed max_limit"
        inputs = {
            "prompt": "count number of patients",
        }
        result = chain.invoke(inputs)
        print("ExecuteTimbrQueryChain with count patients result:", result)
        assert "rows" in result, "Result should contain 'rows'"
        assert isinstance(result["rows"], list), "'rows' should be a list"
        assert result["sql"], "SQL should be present in the result"
        assert "PATIENT" in result["sql"].upper(), "SQL should contain 'PATIENT'"
        assert list(result["rows"][0].values())[0] == 1176837, "Number of rows should not exceed max_limit"

    def test_execute_with_multiple_ontologies4(self, llm, config):
        """Test query execution with multiple ontologies."""
        chain = ExecuteTimbrQueryChain(
            llm=llm,
            url=config["timbr_url"],
            token=config["timbr_token"],
            ontology=[config["timbr_ontology"] , "timbr_patient_journey"],
            concepts_list="None",
            views_list=[config["timbr_ontology"] + "." + "order_cube", "timbr_patient_journey.patient_cube"],
            include_tags="*",
            verify_ssl=config["verify_ssl"],
        )
        inputs = {
            "prompt": "count number of customers",
        }
        result = chain.invoke(inputs)
        print("ExecuteTimbrQueryChain with count customers result:", result)
        assert "rows" in result, "Result should contain 'rows'"
        assert isinstance(result["rows"], list), "'rows' should be a list"
        assert result["sql"], "SQL should be present in the result"
        assert "ORDER_CUBE" in result["sql"].upper(), "SQL should contain 'ORDER_CUBE'"
        assert list(result["rows"][0].values())[0] == 12508, "Number of rows should not exceed max_limit"
        inputs = {
            "prompt": "count number of patients",
        }
        result = chain.invoke(inputs)
        print("ExecuteTimbrQueryChain with count patients result:", result)
        assert "rows" in result, "Result should contain 'rows'"
        assert isinstance(result["rows"], list), "'rows' should be a list"
        assert result["sql"], "SQL should be present in the result"
        assert "PATIENT_CUBE" in result["sql"].upper(), "SQL should contain 'PATIENT_CUBE'"
        assert list(result["rows"][0].values())[0] == 1176837, "Number of rows should not exceed max_limit"